# Copyright 2022 TIT CS.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0

from live4safe_commons import *



from face_recognition import face_locations, face_landmarks
from math import ceil
import base64
import cv2
import ffmpeg
import logging

logger = logging.getLogger("Misc")
logger.setLevel(logging.DEBUG)


class Misc:
    def __init__(self):
        # BBX
        self.locations = None
        self.process_this_frame = None
        self.scale = None
        self.scale_back = None
        self.offset = 0
        # Blink
        self.landmarks = None
        self.eye_left = None
        self.eye_right = None
        self.eye_left_bbx = None
        self.eye_right_bbx = None
        # Mouth
        self.th_area = 0.45
        self.bbx_b_lip = None
        self.bbx_t_lip = None
        self.area = 0
        # Resize
        self.aspect_ratio_4_3 = float(4 / 3)
        self.aspect_ratio_16_9 = float(16 / 9)

        self.process_this_frame = True
        self.scale = 0.25
        self.scale_back = ceil(1 / self.scale)
        self.offset = 20

    def __del__(self):
        pass

    def encodeFrame64(self, frame):
        """Encode frame to base 64.
        @frame: frame CV2_64 to be encoded"""
        if frame is not None:
            ret, buffer = cv2.imencode(".jpg", frame)
            if ret:
                return base64.b64encode(buffer).decode("utf-8")
            return None
        else:
            return None

    def resize_to_VGA(self, frame):
        """Resize a given frame to VGA in 16:9 (854x480) or 4:3 (640X480)
        @frame: Frame to be resized."""
        (h, w) = frame.shape[:2]
        if w / h == self.aspect_ratio_16_9 and (w != 854 or h != 480):
            return cv2.resize(frame, (480, 854), interpolation=cv2.INTER_CUBIC)
        elif w / h == self.aspect_ratio_4_3 and (w != 640 or h != 480):
            return cv2.resize(frame, (480, 640), interpolation=cv2.INTER_CUBIC)
        else:
            return frame

    def resize_to_QVGA(self, frame):
        """Resize a given frame to QVGA in 16:9 (854x480) or 4:3 (640X480)
        @frame: Frame to be resized."""
        (h, w) = frame.shape[:2]
        if float(h / w) == self.aspect_ratio_16_9 and (w != 427 or h != 240):
            return cv2.resize(frame, (240, 426), interpolation=cv2.INTER_CUBIC)
        elif float(h / w) == self.aspect_ratio_4_3 and (w != 320 or h != 240):
            return cv2.resize(frame, (240, 320), interpolation=cv2.INTER_CUBIC)
        else:
            return frame

    def get_BBX(self, frame=None, method=None):
        """Get a face Bounding Box from the videostream frame.
        @frame: frame from the videostream."""
        if frame is not None:
            small_frame = self.resize_to_QVGA(frame)
            # Only process every other frame of video to save time
            if self.process_this_frame:
                rgb_small_frame = small_frame[:, :, ::-1]
                # Find all the faces and face encodings
                # in the current frame of video
                if method is None or method != "multiple_faces":
                    self.locations = face_locations(
                        rgb_small_frame, model="hog")
                else:
                    self.locations = face_locations(
                        rgb_small_frame, model="cnn")

            self.process_this_frame = not self.process_this_frame
            # Initialize face matrix and BBx components that will be returned
            face = []  # np.zeros_like(frame)
            bbx = []
            top, right, bottom, left = (0, 0, 0, 0)

            for (top, right, bottom, left) in self.locations:
                # Scale back up face locations since the frame we
                # detected in was scaled to 1/4 size
                # top *= self.scale_back
                # right *= self.scale_back
                # bottom *= self.scale_back
                # left *= self.scale_back

                # face.append(frame[top:bottom, left:right])
                face.append(small_frame[top:bottom, left:right])
                bbx.append((top, bottom, left, right))
            return face, bbx
        else:
            return None, None

    # FaceLandmarks
    def get_face_landmarks(self, face):
        """Find face landmarks and store given a face frame.
        @face: face frame."""

        self.landmarks = face_landmarks(
            face, face_locations=None, model="large")
        ret_lendmarks = False
        if len(self.landmarks):
            # Left eye
            start = (
                self.landmarks[0]["right_eye"][0][0] - self.offset,
                self.landmarks[0]["right_eye"][0][1] - self.offset,
            )
            end = (
                self.landmarks[0]["right_eye"][3][0] + self.offset,
                self.landmarks[0]["right_eye"][3][1] + self.offset,
            )
            self.eye_left_bbx = (start, end)
            self.eye_right = face[start[1]: end[1], start[0]: end[0]]
            # Right eye
            start = (
                self.landmarks[0]["left_eye"][0][0] - self.offset,
                self.landmarks[0]["left_eye"][0][1] - self.offset,
            )
            end = (
                self.landmarks[0]["left_eye"][3][0] + self.offset,
                self.landmarks[0]["left_eye"][3][1] + self.offset,
            )
            self.eye_right_bbx = (start, end)
            self.eye_left = face[start[1]: end[1], start[0]: end[0]]
            # Mouth
            self.bbx_b_lip = self.landmarks[0]["bottom_lip"]
            self.bbx_t_lip = self.landmarks[0]["top_lip"]
            ret_lendmarks = True
        else:
            self.eye_left = None
            self.eye_right = None
            self.eye_left_bbx = None
            self.eye_right_bbx = None
            self.bbx_b_lip = None
            self.bbx_t_lip = None
        return ret_lendmarks

    # Blink
    def get_eye_left(self):
        """ Returns the left eye matrix frame"""
        return self.eye_left

    def get_eye_right(self):
        """ Returns the right eye matrix frame"""
        return self.eye_right

    def get_eyes_bbx(self):
        """ Returns the right and left eyes bbxs"""
        return (self.eye_right_bbx, self.eye_left_bbx)

    # Mouth
    def get_lips_bbx(self):
        return (self.bbx_t_lip, self.bbx_b_lip)

    def check_face_area(self):
        """ Returns if face area is greather than the minimum required."""
        return self.area > self.th_area

    def compute_face_area(self, bbx, shape):
        """Compute face area given the bbx of a face.
        @bbx: bbx of a detected face.
        @shape: face shape."""
        if len(bbx):
            (y, w, h, x) = bbx
            self.area = (w * h) / (shape[0] * shape[1])

    # https://github.com/melvinkokxw/face_emotion_detection/commit/68febcc3c2e47f723d9fb275be3c0357cf68b31c
    def check_rotation(self, video_file):
        """Checks if given video file has rotation metadata
        Used as workaround for OpenCV bug
        https://github.com/opencv/opencv/issues/15499,
        where the rotation metadata in a video is not used
         by cv2.VideoCapture. May have to be removed/tweaked
        when bug is fixed in a new opencv-python release.
        Parameters
        ----------
        video_file : str
            Path to video file to be checked
        Returns
        -------
        rotate_code : enum or None
            Flag fror cv2.rotate to decide how much to rotate the image.
            None if no rotation is required
        """

        try:
            meta_dict = ffmpeg.probe(video_file)
            rotation_angle = int(meta_dict["streams"][0]["tags"]["rotate"])
        except Exception as ex:
            logger.exception(
                f'Failed to get rotate tag in check_rotation: {video_file}')
            logger.exception(ex)
            return None

        rotate_code = None
        if rotation_angle == 90:
            rotate_code = cv2.ROTATE_90_CLOCKWISE
        elif rotation_angle == 180:
            rotate_code = cv2.ROTATE_180
        elif rotation_angle == 270:
            rotate_code = cv2.ROTATE_90_COUNTERCLOCKWISE

        return rotate_code



import pika
import sys
import os
import argparse
import json
import requests
from pathlib import Path
import logging
from live4safe_commons.liveness_status_enum import LivenessStatusEnum

logger = logging.getLogger()
logger.setLevel(logging.INFO)
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
formatter = logging.Formatter(
    "%(asctime)s - %(name)s - %(levelname)s - %(message)s")
ch.setFormatter(formatter)
logger.addHandler(ch)

UPDATE_URL = os.getenv("UPDATE_URL", "http://localhost:8000/liveness/")
# METHOD = os.getenv("METHOD", "liveness_net")
RABBITMQ_HOST = os.getenv("RABBITMQ_HOST", "127.0.0.1")
RABBITMQ_PORT = os.getenv("RABBITMQ_PORT", "5672")

logger.info("Environment Variables:")
logger.info(UPDATE_URL)
# logger.info(METHOD)
logger.info(RABBITMQ_HOST)
logger.info(RABBITMQ_PORT)

"""
if METHOD == "multiple_faces" or METHOD == "same_face":
    method = __import__("faces.faces", globals(), locals(), [], 0)
    # Same method in faces
    if METHOD == "same_face":
        METHOD = "multiple_faces"
else:
    method = __import__(METHOD + "." + METHOD, globals(), locals(), [], 0)

print(method)
"""


class Connection:
    video_path = None
    port = None
    choosen_method = None
    connection = None
    channel = None
    current_method = None

    def __init__(self, cho_method, cur_method):
        logger.info(f'Initiating {cho_method}...')
        self.choosen_method = cho_method

        try:
            Path("./videos").mkdir(parents=True, exist_ok=True)
            self.current_method = cur_method

        except SystemExit:
            os._exit(0)

    def __del__(self):
        if self.connection is not None:
            self.connection.close()
            del self.connection
            del self.channel
        pth = Path("./videos")
        for file in pth.iterdir():
            file.unlink()
        pth.rmdir()
        del pth

    def set_route(self):
        try:
            logger.info(f'Connecting Pika to {RABBITMQ_HOST}' +
                        'at port {RABBITMQ_PORT}'
                        )
            self.connection = pika.BlockingConnection(
                pika.ConnectionParameters(
                    host=RABBITMQ_HOST, port=RABBITMQ_PORT)
            )
            self.channel = self.connection.channel()
            self.channel.queue_declare(queue=self.choosen_method)
            # self.channel.queue_declare(queue=f"{queue_name}.results")
        except Exception as ex:
            logger.error(f'Connecting Pika to {RABBITMQ_HOST}' +
                         'at port {RABBITMQ_PORT}'
                         )
            logger.exception(ex)
            logger.critical("Aborting...")
            exit()

    def callback(self, ch, method, properties, body):
        logger.info("Method: {}".format(self.choosen_method))
        logger.info("Analysing Video: {}".format(body.decode("utf-8")))
        message = json.loads(body.decode("utf-8"))
        answer = {}

        try:
            # Download video file from video_path in Google Storage
            logger.info("Downloading video from Google Storage")
            pth = "./videos/" + str(message["video_path"]).split("/")[-1]
            header = {"Authorization": message.get("token")}
            r = requests.get(message["video_path"], allow_redirects=True)
            if r.ok:
                with open(pth, "wb") as target:
                    target.write(r.content)
                # PROCESS VIDEO
                requests.patch(
                    url=UPDATE_URL+message.get("liveness_id")+"/",
                    headers=header,
                    json={
                        f'{self.choosen_method}_status:' +
                        LivenessStatusEnum.processing.name},
                )
                logger.info("Starting Liveness method: {}".format(
                    self.choosen_method))
                answer = self.current_method.analyseVideo(pth)

                print(answer)
                logger.info(
                    "Finishing Liveness method: {}, ANSWER: {}"
                    .format(self.choosen_method, answer))

                # Delete downloaded file after processing
                Path(pth).unlink()
                requests.patch(url=UPDATE_URL +
                               message.get("liveness_id")+"/",
                               headers=header,
                               json=answer)
            else:
                logger.error(
                    "Video %s was not found on server, aborting processing...",
                    str(message["video_path"]).split("/")[-1],
                )
                requests.put(
                    url=UPDATE_URL + message.get("liveness_id"),
                    json={
                        f"{self.choosen_method}_status":
                        LivenessStatusEnum.error.name},
                )

        except Exception as ex:
            logger.exception(
                "An exception ocurred when trying to process video: %s",
                ex
            )
            requests.put(
                url=UPDATE_URL + message.get("liveness_id")+"/",
                json={
                    f"{self.choosen_method}_status":
                    LivenessStatusEnum.execption.name},
            )

    def get_request(self):
        try:
            # os.system("cls" if os.name == "nt" else "clear")
            logger.info("Setting Route")
            self.set_route()

            logger.info("Start consuming")
            self.channel.basic_consume(
                queue=self.choosen_method,
                on_message_callback=self.callback,
                auto_ack=True,
            )
            print("\n\n[*] Waiting for messages. To exit press CTRL+C")
            self.channel.start_consuming()
        except KeyboardInterrupt:
            logger.info("Interrupted")
            try:
                sys.exit(0)
            except SystemExit:
                os._exit(0)


"""
def main():
    con = Connection()
    con.get_request()


if __name__ == "__main__":
    ap = argparse.ArgumentParser()
    ap.add_argument(
        "-m",
        "--method",
        type=str,
        default=METHOD,
        help="Method analysed in liveness.",
    )
    args = vars(ap.parse_args())
    logger.info(args)
    main()
"""



from os import getenv
from google.cloud import storage
from azure.storage.blob import BlobServiceClient, ContentSettings, __version__
from azure.core import exceptions
import logging

logger = logging.getLogger("cloud_storage")
logger.setLevel(logging.DEBUG)
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
formatter = logging.Formatter(
    "%(asctime)s - %(name)s - %(levelname)s - %(message)s")
ch.setFormatter(formatter)
logger.addHandler(ch)

CLOUD_STORAGE_TYPE = getenv("CLOUD_STORAGE_TYPE", "AZURE")  # ['GCP', 'AZURE']
STORAGE_PATH = getenv("STORAGE_PATH", "liveness-dev/videos/")
BUCKET_NAME = getenv("BUCKET_NAME", "novaweb-storage")
AZURE_STORAGE_CONNECTION_STRING = getenv('AZURE_STORAGE_CONNECTION_STRING')
AZURE_STORAGE_ACCOUNT = getenv('AZURE_STORAGE_ACCOUNT', "sainovacao")


class CloudStorage:
    def upload_blob(
        self,
        source_file_name,
        destination_blob_name: str,
        blob_type: str,
    ):
        pass

    def delete_blob(self, blob_name: str):
        pass

    def get_url_prefix(self) -> str:
        pass


class GoogleStorage(CloudStorage):
    def __init__(self) -> None:
        super().__init__()
        self.storage_path = STORAGE_PATH
        self.bucket_name = BUCKET_NAME

    def upload_blob(
        self,
        source_file_name,
        destination_blob_name: str,
        blob_type: str = "video/mp4",
    ):
        """Uploads a file to the bucket."""

        storage_client = storage.Client()
        bucket = storage_client.bucket(BUCKET_NAME)
        blob = bucket.blob(STORAGE_PATH + destination_blob_name)

        # blob.upload_from_filename(source_file_name)
        blob.upload_from_file(source_file_name, content_type=blob_type)

        logger.info(
            "File {} uploaded to {}.".format(
                source_file_name, STORAGE_PATH + destination_blob_name
            )
        )

    def delete_blob(self, blob_name: str):
        """Deletes a blob from the bucket."""
        blob_name = blob_name.replace(
            "https://storage.googleapis.com/" + self.bucket_name + "/", ""
        )

        storage_client = storage.Client()

        bucket = storage_client.bucket(BUCKET_NAME)
        blob = bucket.blob(blob_name)
        blob.delete()

        logger.info("Blob {} deleted.".format(blob_name))

    def get_url_prefix(self) -> str:
        return (
            "https://storage.googleapis.com/"
            + self.bucket_name
            + "/"
            + self.storage_path
            if self.storage_path.endswith("/")
            else self.storage_path + "/"
        )


class AzureStorage(CloudStorage):
    def __init__(self) -> None:
        super().__init__()
        try:
            logger.info("Azure Blob Storage v" + __version__)

            # Create the BlobServiceClient object which
            # will be used to create a container client
            self.blob_service_client = BlobServiceClient.from_connection_string(
                AZURE_STORAGE_CONNECTION_STRING)

            # Create the container
            self.container_client = self.blob_service_client.create_container(
                BUCKET_NAME)

        except exceptions.ResourceExistsError as ex:
            logger.info(
                f"Container '{BUCKET_NAME}' already exists. Exception: {ex}")

        except Exception as ex:
            logger.exception(f'General exception when create bucket: {ex}')

    def upload_blob(
        self,
        source_file_name,
        destination_blob_name: str,
        blob_type: str,
    ):
        # Create a blob client using the
        # local file name as the name for the blob
        blob_client = self.blob_service_client.get_blob_client(
            container=BUCKET_NAME, blob=destination_blob_name)

        # Upload the created file
        content_settings = ContentSettings(content_type=blob_type)
        blob_client.upload_blob(source_file_name.read(),
                                content_settings=content_settings)

        logger.info(
            "File {} uploaded to container {}: {}".format(
                source_file_name, BUCKET_NAME, destination_blob_name
            )
        )

    def delete_blob(self, blob_name: str):
        blob_name = blob_name.replace(
            self.get_url_prefix(), ""
        )

        # Create a blob client using the
        # local file name as the name for the blob
        blob_client = self.blob_service_client.get_blob_client(
            container=BUCKET_NAME, blob=blob_name)

        # Delete the file
        blob_client.delete_blob()

        logger.info("Blob {} deleted.".format(blob_name))

    def get_url_prefix(self) -> str:
        return (
            f"https://{AZURE_STORAGE_ACCOUNT}." +
            "blob.core.windows.net/{BUCKET_NAME}/"
        )


def get_storage():
    if CLOUD_STORAGE_TYPE == "GCP":
        return GoogleStorage()
    elif CLOUD_STORAGE_TYPE == "AZURE":
        return AzureStorage()
    else:
        logger.critical(
            f"CLOUD_STORAGE_TYPE not recognized: {CLOUD_STORAGE_TYPE}"
        )
        raise Exception("CLOUD_STORAGE_TYPE not recognized")



from setuptools import setup

setup(
    name='life4safe-commons',
    version='0.0.2',
    author='TIT CS',
    author_email='contato@titcs.com.br',
    packages=['live4safe_commons'],
    description='Common functions to liveliness algorithms',
    long_description='Common functions to liveliness algorithms',
    url='https://TITBrasil@dev.azure.com/TITBrasil/TITCS%20-%20Live4Safe/_git/Live4Safe_Common',
    project_urls={
        'Código fonte': 'https://TITBrasil@dev.azure.com/TITBrasil/TITCS%20-%20Live4Safe/_git/Live4Safe_Common',
        'Download': 'https://dev.azure.com/TITBrasil/TITCS%20-%20Live4Safe/_git/Live4Safe_Common'
    },
    license='MIT',
    keywords='live4safe',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: Portuguese (Brazilian)',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Internationalization',
        'Topic :: Scientific/Engineering :: Physics'
    ]
)



from enum import Enum


class LivenessStatusEnum(Enum):
    unknown = 0
    not_requested = 1
    queued = 10
    pending = 11
    processing = 20
    processed = 21
    error = 40
    execption = 50
